/*
 * This file is part of ltrace.
 * Copyright (C) 2022 Kai Zhang (laokz)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
/* from linux kernel 5.10 */
    "SIG_0",           /* 0 */
    "SIGHUP",          /* 1 */
    "SIGINT",          /* 2 */
    "SIGQUIT",         /* 3 */
    "SIGILL",          /* 4 */
    "SIGTRAP",         /* 5 */
    "SIGABRT",         /* 6 */
    "SIGBUS",          /* 7 */
    "SIGFPE",          /* 8 */
    "SIGKILL",         /* 9 */
    "SIGUSR1",         /* 10 */
    "SIGSEGV",         /* 11 */
    "SIGUSR2",         /* 12 */
    "SIGPIPE",         /* 13 */
    "SIGALRM",         /* 14 */
    "SIGTERM",         /* 15 */
    "SIGSTKFLT",       /* 16 */
    "SIGCHLD",         /* 17 */
    "SIGCONT",         /* 18 */
    "SIGSTOP",         /* 19 */
    "SIGTSTP",         /* 20 */
    "SIGTTIN",         /* 21 */
    "SIGTTOU",         /* 22 */
    "SIGURG",          /* 23 */
    "SIGXCPU",         /* 24 */
    "SIGXFSZ",         /* 25 */
    "SIGVTALRM",       /* 26 */
    "SIGPROF",         /* 27 */
    "SIGWINCH",        /* 28 */
    "SIGIO",           /* 29 */
    "SIGPWR",          /* 30 */
    "SIGSYS",          /* 31 */
