/*
 * This file is part of ltrace.
 * Copyright (C) 2010 Arnaud Patard, Mandriva SA
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

	"SIG_0",           /* 0 */
	"SIGHUP",          /* 1 */
	"SIGINT",          /* 2 */
	"SIGQUIT",         /* 3 */
	"SIGILL",          /* 4 */
	"SIGTRAP",         /* 5 */
	"SIGIOT",          /* 6 */
	"SIGEMT",          /* 7 */
	"SIGFPE",          /* 8 */
	"SIGKILL",         /* 9 */
	"SIGBUS",          /* 10 */
	"SIGSEGV",         /* 11 */
	"SIGSYS",          /* 12 */
	"SIGPIPE",         /* 13 */
	"SIGALRM",         /* 14 */
	"SIGTERM",         /* 15 */
	"SIGUSR1",         /* 16 */
	"SIGUSR2",         /* 17 */
	"SIGCHLD",         /* 18 */
	"SIGPWR",          /* 19 */
	"SIGWINCH",        /* 20 */
	"SIGURG",          /* 21 */
	"SIGIO",           /* 22 */
	"SIGSTOP",         /* 23 */
	"SIGTSTP",         /* 24 */
	"SIGCONT",         /* 25 */
	"SIGTTIN",         /* 26 */
	"SIGTTOU",         /* 27 */
	"SIGVTALRM",       /* 28 */
	"SIGPROF",         /* 29 */
	"SIGXCPU",         /* 30 */
	"SIGXFSZ",         /* 31 */
