/*
 * This file is part of ltrace.
 * Copyright (C) 2012,2013 Petr Machata, Red Hat Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

/* Important types defined in other header files are declared
   here.  */
struct Event;
struct arg_type_info;
struct breakpoint;
struct dict;
struct expr_node;
struct filter;
struct library;
struct library_symbol;
struct ltelf;
struct param;
struct param_enum;
struct process;
struct protolib;
struct protolib_cache;
struct prototype;
struct timedelta;
struct value;
struct value_dict;
struct vect;
