/*
 * This file is part of ltrace.
 * Copyright (C) 1998,2008 Juan Cespedes
 * Copyright (C) 2006 Ian Wienand
 * Copyright (C) 2006 Steve Fink
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#ifndef DEFAULT_ALIGN
#define DEFAULT_ALIGN 50	/* default alignment column for results */
#endif				/* (-a switch) */

#ifndef DEFAULT_STRLEN
#define DEFAULT_STRLEN  32	/* default maximum # of bytes printed in */
#endif				/* strings (-s switch) */

#ifndef DEFAULT_ARRAYLEN
#define DEFAULT_ARRAYLEN  4	/* default maximum # array elements */
#endif				/* (-A switch) */
