/*
 * This file is part of ltrace.
 * Copyright (C) 2013 Petr Machata, Red Hat Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include "bits.h"

/* This is called rarely, and any overhead will be lost in ptrace
 * noise, so the algorithm doesn't need to be terribly clever.  For
 * the same reason we don't bother defining the corresponding _32
 * variant.  */
unsigned
bitcount(uint64_t u)
{
	int c = 0;
	for (; u > 0; u &= u - 1)
		c++;
	return c;
}
